@@ -5,7 +5,7 @@ from logging import warning
 from json import JSONEncoder
 from sys import version
 from decimal import Decimal
-from .utils import hashodict, call_with_optional_kwargs
+from .utils import hashodict, call_with_optional_kwargs, NoPandasException, NoNumpyException
 
 
 class TricksEncoder(JSONEncoder):
@@ -42,6 +42,7 @@ class TricksEncoder(JSONEncoder):
 		prev_id = id(obj)
 		for encoder in self.obj_encoders:
 			if hasattr(encoder, 'default'):
+				#todo: write test for this scenario (maybe ClassInstanceEncoder?)
 				try:
 					obj = call_with_optional_kwargs(encoder.default, obj, primitives=self.primitives)
 				except TypeError as err:
@@ -52,6 +53,7 @@ class TricksEncoder(JSONEncoder):
 			else:
 				raise TypeError('`obj_encoder` {0:} does not have `default` method and is not callable'.format(encoder))
 		if id(obj) == prev_id:
+			#todo: test
 			raise TypeError('Object of type {0:} could not be encoded by {1:} using encoders [{2:s}]'.format(
 				type(obj), self.__class__.__name__, ', '.join(str(encoder) for encoder in self.obj_encoders)))
 		return obj
@@ -100,7 +102,7 @@ def class_instance_encode(obj, primitives=False):
 	"""
 	if isinstance(obj, list) or isinstance(obj, dict):
 		return obj
-	if hasattr(obj, '__class__') and hasattr(obj, '__dict__'):
+	if hasattr(obj, '__class__') and (hasattr(obj, '__dict__') or hasattr(obj, '__slots__')):
 		if not hasattr(obj, '__new__'):
 			raise TypeError('class "{0:s}" does not have a __new__ method; '.format(obj.__class__) +
 				('perhaps it is an old-style class not derived from `object`; add `object` as a base class to encode it.'
@@ -108,7 +110,7 @@ def class_instance_encode(obj, primitives=False):
 		try:
 			obj.__new__(obj.__class__)
 		except TypeError:
-			raise TypeError(('instance "{0:s}" of class "{1:s}" cannot be encoded because it\'s __new__ method '
+			raise TypeError(('instance "{0:}" of class "{1:}" cannot be encoded because it\'s __new__ method '
 				'cannot be called, perhaps it requires extra parameters').format(obj, obj.__class__))
 		mod = obj.__class__.__module__
 		if mod == '__main__':
@@ -119,12 +121,27 @@ def class_instance_encode(obj, primitives=False):
 		name = obj.__class__.__name__
 		if hasattr(obj, '__json_encode__'):
 			attrs = obj.__json_encode__()
-		else:
-			attrs = hashodict(obj.__dict__.items())
+			if primitives:
+				return attrs
+			else:
+				return hashodict((('__instance_type__', (mod, name)), ('attributes', attrs)))
+		dct = hashodict([('__instance_type__',(mod, name))])
+		if hasattr(obj, '__slots__'):
+			slots = obj.__slots__
+			if isinstance(slots, str):
+				slots = [slots]
+			slots = list(item for item in slots if item != '__dict__')
+			dct['slots'] = hashodict([])
+			for s in slots:
+				dct['slots'][s] = getattr(obj, s)
+		if hasattr(obj, '__dict__'):
+			dct['attributes'] = hashodict(obj.__dict__)
 		if primitives:
+			attrs = dct.get('attributes',{})
+			attrs.update(dct.get('slots',{}))
 			return attrs
 		else:
-			return hashodict((('__instance_type__', (mod, name)), ('attributes', attrs)))
+			return dct
 	return obj
 
 
@@ -172,6 +189,7 @@ class ClassInstanceEncoder(JSONEncoder):
 	"""
 	See `class_instance_encoder`.
 	"""
+	# Not covered in tests since `class_instance_encode` is recommended way.
 	def __init__(self, obj, encode_cls_instances=True, **kwargs):
 		self.encode_cls_instances = encode_cls_instances
 		super(ClassInstanceEncoder, self).__init__(obj, **kwargs)
@@ -200,3 +218,109 @@ def json_set_encode(obj, primitives=False):
 	return obj
 
 
+def pandas_encode(obj, primitives=False):
+	from pandas import DataFrame, Series
+	if isinstance(obj, (DataFrame, Series)):
+		#todo: this is experimental
+		if not getattr(pandas_encode, '_warned', False):
+			pandas_encode._warned = True
+			warning('Pandas dumping support in json-tricks is experimental and may change in future versions.')
+	if isinstance(obj, DataFrame):
+		repr = hashodict()
+		if not primitives:
+			repr['__pandas_dataframe__'] = hashodict((
+				('column_order', tuple(obj.columns.values)),
+				('types', tuple(str(dt) for dt in obj.dtypes)),
+			))
+		repr['index'] = tuple(obj.index.values)
+		for k, name in enumerate(obj.columns.values):
+			repr[name] = tuple(obj.ix[:, k].values)
+		return repr
+	if isinstance(obj, Series):
+		repr = hashodict()
+		if not primitives:
+			repr['__pandas_series__'] = hashodict((
+				('name', str(obj.name)),
+				('type', str(obj.dtype)),
+			))
+		repr['index'] = tuple(obj.index.values)
+		repr['data'] = tuple(obj.values)
+		return repr
+	return obj
+
+
+def nopandas_encode(obj):
+	if ('DataFrame' in getattr(obj.__class__, '__name__', '') or 'Series' in getattr(obj.__class__, '__name__', '')) \
+			and 'pandas.' in getattr(obj.__class__, '__module__', ''):
+		raise NoPandasException(('Trying to encode an object of type {0:} which appears to be '
+			'a numpy array, but numpy support is not enabled, perhaps it is not installed.').format(type(obj)))
+	return obj
+
+
+def numpy_encode(obj, primitives=False):
+	"""
+	Encodes numpy `ndarray`s as lists with meta data.
+	
+	Encodes numpy scalar types as Python equivalents. Special encoding is not possible,
+	because int64 (in py2) and float64 (in py2 and py3) are subclasses of primitives,
+	which never reach the encoder.
+	
+	:param primitives: If True, arrays are serialized as (nested) lists without meta info.
+	"""
+	from numpy import ndarray, generic
+	if isinstance(obj, ndarray):
+		if primitives:
+			return obj.tolist()
+		else:
+			dct = hashodict((
+				('__ndarray__', obj.tolist()),
+				('dtype', str(obj.dtype)),
+				('shape', obj.shape),
+			))
+			if len(obj.shape) > 1:
+				dct['Corder'] = obj.flags['C_CONTIGUOUS']
+			return dct
+	elif isinstance(obj, generic):
+		if NumpyEncoder.SHOW_SCALAR_WARNING:
+			NumpyEncoder.SHOW_SCALAR_WARNING = False
+			warning('json-tricks: numpy scalar serialization is experimental and may work differently in future versions')
+		return obj.item()
+	return obj
+
+
+class NumpyEncoder(ClassInstanceEncoder):
+	"""
+	JSON encoder for numpy arrays.
+	"""
+	SHOW_SCALAR_WARNING = True  # show a warning that numpy scalar serialization is experimental
+	
+	def default(self, obj, *args, **kwargs):
+		"""
+		If input object is a ndarray it will be converted into a dict holding
+		data type, shape and the data. The object can be restored using json_numpy_obj_hook.
+		"""
+		warning('`NumpyEncoder` is deprecated, use `numpy_encode`')  #todo
+		obj = numpy_encode(obj)
+		return super(NumpyEncoder, self).default(obj, *args, **kwargs)
+
+
+def nonumpy_encode(obj):
+	"""
+	Raises an error for numpy arrays.
+	"""
+	if 'ndarray' in getattr(obj.__class__, '__name__', '') and 'numpy.' in getattr(obj.__class__, '__module__', ''):
+		raise NoNumpyException(('Trying to encode an object of type {0:} which appears to be '
+			'a pandas data stucture, but pandas support is not enabled, perhaps it is not installed.').format(type(obj)))
+	return obj
+
+
+class NoNumpyEncoder(JSONEncoder):
+	"""
+	See `nonumpy_encode`.
+	"""
+	def default(self, obj, *args, **kwargs):
+		warning('`NoNumpyEncoder` is deprecated, use `nonumpy_encode`')  #todo
+		obj = nonumpy_encode(obj)
+		return super(NoNumpyEncoder, self).default(obj, *args, **kwargs)
+
+
