@@ -4,63 +4,52 @@ from io import BytesIO
 from json import loads as json_loads
 from os import fsync
 from sys import exc_info, version
+from .utils import NoNumpyException  # keep 'unused' imports
 from .comment import strip_comment_line_with_symbol, strip_comments  # keep 'unused' imports
 from .encoders import TricksEncoder, json_date_time_encode, class_instance_encode, ClassInstanceEncoder, \
-	json_complex_encode, json_set_encode, numeric_types_encode  # keep 'unused' imports
+	json_complex_encode, json_set_encode, numeric_types_encode, numpy_encode, nonumpy_encode, NoNumpyEncoder, \
+	nopandas_encode, pandas_encode  # keep 'unused' imports
 from .decoders import DuplicateJsonKeyException, TricksPairHook, json_date_time_hook, ClassInstanceHook, \
-	json_complex_hook, json_set_hook, numeric_types_hook  # keep 'unused' imports
+	json_complex_hook, json_set_hook, numeric_types_hook, json_numpy_obj_hook, json_nonumpy_obj_hook, \
+	nopandas_hook, pandas_hook  # keep 'unused' imports
 from json import JSONEncoder
 
 
 is_py3 = (version[:2] == '3.')
-str_type = str if is_py3 else basestring
+str_type = str if is_py3 else (basestring, unicode,)
 ENCODING = 'UTF-8'
 
 
-class NoNumpyException(Exception):
-		""" Trying to use numpy features, but numpy cannot be found. """
-
-
-def nonumpy_encode(obj, primitives=False):
-	"""
-	Emits a warning for numpy arrays, no other effect.
-	"""
-	if 'ndarray' in str(type(obj)):
-		raise NoNumpyException(('Trying to encode {0:} which appears to be a numpy array ({1:}), but numpy ' +
-			'support is not enabled. Make sure that numpy is installed and that you import from json_tricks.np.')
-			.format(obj, type(obj)))
-	return obj
-
-
-class NoNumpyEncoder(JSONEncoder):
-	"""
-	See `nonumpy_encoder`.
-	"""
-	def default(self, obj, *args, **kwargs):
-		obj = nonumpy_encode(obj)
-		return super(NoNumpyEncoder, self).default(obj, *args, **kwargs)
-
-
-def json_nonumpy_obj_hook(dct):
-	"""
-	This hook has no effect except to check if you're trying to decode numpy arrays without support, and give you a useful message.
-	"""
-	if isinstance(dct, dict) and '__ndarray__' in dct:
-		raise NoNumpyException(('Trying to decode dictionary ({0:}) which appears to be a numpy array, but numpy ' +
-			'support is not enabled. Make sure that numpy is installed and that you import from json_tricks.np.')
-			.format(', '.join(dct.keys()[:10])))
-	return dct
-
-
 _cih_instance = ClassInstanceHook()
-DEFAULT_ENCODERS = (json_date_time_encode, class_instance_encode, json_complex_encode, json_set_encode, numeric_types_encode,)
-DEFAULT_NONP_ENCODERS = (nonumpy_encode,) + DEFAULT_ENCODERS
-DEFAULT_HOOKS = (json_date_time_hook, _cih_instance, json_complex_hook, json_set_hook, numeric_types_hook,)
-DEFAULT_NONP_HOOKS = (json_nonumpy_obj_hook,) + DEFAULT_HOOKS
+DEFAULT_ENCODERS = [json_date_time_encode, json_complex_encode, json_set_encode, numeric_types_encode, class_instance_encode,]
+DEFAULT_HOOKS = [json_date_time_hook, json_complex_hook, json_set_hook, numeric_types_hook, _cih_instance,]
+
+try:
+	import numpy
+except ImportError:
+	DEFAULT_ENCODERS = [nonumpy_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [json_nonumpy_obj_hook,] + DEFAULT_HOOKS
+else:
+	# numpy encode needs to be before complex
+	DEFAULT_ENCODERS = [numpy_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [json_numpy_obj_hook,] + DEFAULT_HOOKS
+
+try:
+	import pandas
+except ImportError:
+	DEFAULT_ENCODERS = [nopandas_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [nopandas_hook,] + DEFAULT_HOOKS
+else:
+	DEFAULT_ENCODERS = [pandas_encode,] + DEFAULT_ENCODERS
+	DEFAULT_HOOKS = [pandas_hook,] + DEFAULT_HOOKS
 
 
-def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_NONP_ENCODERS, extra_obj_encoders=(),
-		primitives=False, compression=None, allow_nan=False, **jsonkwargs):
+DEFAULT_NONP_ENCODERS = [nonumpy_encode,] + DEFAULT_ENCODERS    # DEPRECATED
+DEFAULT_NONP_HOOKS = [json_nonumpy_obj_hook,] + DEFAULT_HOOKS   # DEPRECATED
+
+
+def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS, extra_obj_encoders=(),
+		primitives=False, compression=None, allow_nan=False, conv_str_byte=False, **jsonkwargs):
 	"""
 	Convert a nested data structure to a json string.
 
@@ -70,30 +59,32 @@ def dumps(obj, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_NONP_ENCO
 	:param obj_encoders: Iterable of encoders to use to convert arbitrary objects into json-able promitives.
 	:param extra_obj_encoders: Like `obj_encoders` but on top of them: use this to add encoders without replacing defaults. Since v3.5 these happen before default encoders.
 	:param allow_nan: Allow NaN and Infinity values, which is a (useful) violation of the JSON standard (default False).
+	:param conv_str_byte: Try to automatically convert between strings and bytes (assuming utf-8) (default False).
 	:return: The string containing the json-encoded version of obj.
 
 	Other arguments are passed on to `cls`. Note that `sort_keys` should be false if you want to preserve order.
-
-	Use `json_tricks.np.dumps` instead if you want encoding of numpy arrays.
 	"""
+	if not hasattr(extra_obj_encoders, '__iter__'):
+		raise TypeError('`extra_obj_encoders` should be a tuple in `json_tricks.dump(s)`')
 	encoders = tuple(extra_obj_encoders) + tuple(obj_encoders)
-	string = cls(sort_keys=sort_keys, obj_encoders=encoders, allow_nan=allow_nan,
+	txt = cls(sort_keys=sort_keys, obj_encoders=encoders, allow_nan=allow_nan,
 		primitives=primitives, **jsonkwargs).encode(obj)
+	if not is_py3 and isinstance(txt, str):
+		txt = unicode(txt, ENCODING)
 	if not compression:
-		return string
+		return txt
 	if compression is True:
 		compression = 5
-	if is_py3:
-		string = bytes(string, encoding=ENCODING)
+	txt = txt.encode(ENCODING)
 	sh = BytesIO()
 	with GzipFile(mode='wb', fileobj=sh, compresslevel=compression) as zh:
-		zh.write(string)
+		zh.write(txt)
 	gzstring = sh.getvalue()
 	return gzstring
 
 
-def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_NONP_ENCODERS, extra_obj_encoders=(),
-         primitives=False, compression=None, force_flush=False, allow_nan=False, **jsonkwargs):
+def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_ENCODERS, extra_obj_encoders=(),
+		 primitives=False, compression=None, force_flush=False, allow_nan=False, conv_str_byte=False, **jsonkwargs):
 	"""
 	Convert a nested data structure to a json string.
 
@@ -102,18 +93,33 @@ def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_NONP_E
 	:param force_flush: If True, flush the file handle used, when possibly also in the operating system (default False).
 	
 	The other arguments are identical to `dumps`.
-
-	Use `json_tricks.np.dump` instead if you want encoding of numpy arrays.
 	"""
-	string = dumps(obj, sort_keys=sort_keys, cls=cls, obj_encoders=obj_encoders, extra_obj_encoders=extra_obj_encoders,
-		primitives=primitives, compression=compression, allow_nan=allow_nan, **jsonkwargs)
+	txt = dumps(obj, sort_keys=sort_keys, cls=cls, obj_encoders=obj_encoders, extra_obj_encoders=extra_obj_encoders,
+		primitives=primitives, compression=compression, allow_nan=allow_nan, conv_str_byte=conv_str_byte, **jsonkwargs)
 	if isinstance(fp, str_type):
 		fh = open(fp, 'wb+')
 	else:
 		fh = fp
+		if conv_str_byte:
+			try:
+				fh.write(b'')
+			except TypeError:
+				pass
+				# if not isinstance(txt, str_type):
+				# 	# Cannot write bytes, so must be in text mode, but we didn't get a text
+				# 	if not compression:
+				# 		txt = txt.decode(ENCODING)
+			else:
+				try:
+					fh.write(u'')
+				except TypeError:
+					if isinstance(txt, str_type):
+						txt = txt.encode(ENCODING)
 	try:
+		if 'b' not in getattr(fh, 'mode', 'b?') and not isinstance(txt, str_type) and compression:
+			raise IOError('If compression is enabled, the file must be opened in binary mode.')
 		try:
-			fh.write(string)
+			fh.write(txt)
 		except TypeError as err:
 			err.args = (err.args[0] + '. A possible reason is that the file is not opened in binary mode; '
 				'be sure to set file mode to something like "wb".',)
@@ -128,11 +134,11 @@ def dump(obj, fp, sort_keys=None, cls=TricksEncoder, obj_encoders=DEFAULT_NONP_E
 				pass
 		if isinstance(fp, str_type):
 			fh.close()
-	return string
+	return txt
 
 
-def loads(string, preserve_order=True, ignore_comments=True, decompression=None, obj_pairs_hooks=DEFAULT_NONP_HOOKS,
-		extra_obj_pairs_hooks=(), cls_lookup_map=None, allow_duplicates=True, **jsonkwargs):
+def loads(string, preserve_order=True, ignore_comments=True, decompression=None, obj_pairs_hooks=DEFAULT_HOOKS,
+		extra_obj_pairs_hooks=(), cls_lookup_map=None, allow_duplicates=True, conv_str_byte=False, **jsonkwargs):
 	"""
 	Convert a nested data structure to a json string.
 
@@ -140,25 +146,32 @@ def loads(string, preserve_order=True, ignore_comments=True, decompression=None,
 	:param decode_cls_instances: True to attempt to decode class instances (requires the environment to be similar the the encoding one).
 	:param preserve_order: Whether to preserve order by using OrderedDicts or not.
 	:param ignore_comments: Remove comments (starting with # or //).
-	:param decompression: True to use gzip decompression, False to use raw data, None to automatically determine (default).
+	:param decompression: True to use gzip decompression, False to use raw data, None to automatically determine (default). Assumes utf-8 encoding!
 	:param obj_pairs_hooks: A list of dictionary hooks to apply.
 	:param extra_obj_pairs_hooks: Like `obj_pairs_hooks` but on top of them: use this to add hooks without replacing defaults. Since v3.5 these happen before default hooks.
 	:param cls_lookup_map: If set to a dict, for example ``globals()``, then classes encoded from __main__ are looked up this dict.
 	:param allow_duplicates: If set to False, an error will be raised when loading a json-map that contains duplicate keys.
 	:param parse_float: A function to parse strings to integers (e.g. Decimal). There is also `parse_int`.
+	:param conv_str_byte: Try to automatically convert between strings and bytes (assuming utf-8) (default False).
 	:return: The string containing the json-encoded version of obj.
 
 	Other arguments are passed on to json_func.
-
-	Use json_tricks.np.loads instead if you want decoding of numpy arrays.
 	"""
+	if not hasattr(extra_obj_pairs_hooks, '__iter__'):
+		raise TypeError('`extra_obj_pairs_hooks` should be a tuple in `json_tricks.load(s)`')
 	if decompression is None:
-		decompression = string[:2] == b'\x1f\x8b'
+		decompression = isinstance(string, bytes) and string[:2] == b'\x1f\x8b'
 	if decompression:
 		with GzipFile(fileobj=BytesIO(string), mode='rb') as zh:
 			string = zh.read()
-			if is_py3:
-				string = str(string, encoding=ENCODING)
+			string = string.decode(ENCODING)
+	if not isinstance(string, str_type):
+		if conv_str_byte:
+			string = string.decode(ENCODING)
+		else:
+			raise TypeError(('Cannot automatically encode object of type "{0:}" in `json_tricks.load(s)` since '
+				'the encoding is not known. You should instead encode the bytes to a string and pass that '
+				'string to `load(s)`, for example bytevar.encode("utf-8") if utf-8 is the encoding.').format(type(string)))
 	if ignore_comments:
 		string = strip_comments(string)
 	obj_pairs_hooks = tuple(obj_pairs_hooks)
@@ -168,16 +181,14 @@ def loads(string, preserve_order=True, ignore_comments=True, decompression=None,
 	return json_loads(string, object_pairs_hook=hook, **jsonkwargs)
 
 
-def load(fp, preserve_order=True, ignore_comments=True, decompression=None, obj_pairs_hooks=DEFAULT_NONP_HOOKS,
-		extra_obj_pairs_hooks=(), cls_lookup_map=None, allow_duplicates=True, **jsonkwargs):
+def load(fp, preserve_order=True, ignore_comments=True, decompression=None, obj_pairs_hooks=DEFAULT_HOOKS,
+		extra_obj_pairs_hooks=(), cls_lookup_map=None, allow_duplicates=True, conv_str_byte=False, **jsonkwargs):
 	"""
 	Convert a nested data structure to a json string.
 
 	:param fp: File handle or path to load from.
 
 	The other arguments are identical to loads.
-
-	Use json_tricks.np.load instead if you want decoding of numpy arrays.
 	"""
 	try:
 		if isinstance(fp, str_type):
@@ -186,10 +197,11 @@ def load(fp, preserve_order=True, ignore_comments=True, decompression=None, obj_
 		else:
 			string = fp.read()
 	except UnicodeDecodeError as err:
+		# todo: not covered in tests, is it relevant?
 		raise Exception('There was a problem decoding the file content. A possible reason is that the file is not ' +
 			'opened  in binary mode; be sure to set file mode to something like "rb".').with_traceback(exc_info()[2])
 	return loads(string, preserve_order=preserve_order, ignore_comments=ignore_comments, decompression=decompression,
 		obj_pairs_hooks=obj_pairs_hooks, extra_obj_pairs_hooks=extra_obj_pairs_hooks, cls_lookup_map=cls_lookup_map,
-		allow_duplicates=allow_duplicates, **jsonkwargs)
+		allow_duplicates=allow_duplicates, conv_str_byte=conv_str_byte, **jsonkwargs)
 
 
